/* $Id: slist.cc,v 1.14 1999/02/04 06:10:18 sean Exp sean $ */

//
//  Three parts to this file:  Creaters (make class object)
//			       Functions (the meaty part)
//			       Printers (the debug print functions)
//
//
// Overall process from triangles to scanlines to framebuffer:
//
//          triangle list
//		|
//		| [get_scanlines]
//		V
//	    scanline segments forall scanlines
//		|
//		| [emit_segment]
//		V
//	   unsorted segment list foreach scanline
//		|
//		| [heap sort by segment enpoints]
//		V
//	   endpoint array
//		|
//		| [kahuna]
//		V
//	   sorted segment list foreach scanline
//		|
//		| [texture lighting engine]
//		V
//	   screen buffer
//-----------------------------------------------------

#include <stdio.h>  /* for print operations */
#include <malloc.h> /* for list allocation */
#include "slist.h"
#include "myvga.h"

#define FIXED_SHIFT_EXP 8
#define Z_CLOSE 0.01

///
/// Debug crud:
///

void emit_stub(int, int, int, int);

///////
/////// Creaters:
///////


sl_screen::sl_screen(int ix, int iy)
{
	int y;
	sl_scanline *sc, *nsc;
	sl_screen *screen;
	sl_node *node;

	xres = ix;
	yres = iy;
	sl_node_count = 0;
	sl_used_count = 0;
	endpoint_list_length = 0;
	consideration_list_length = 0;
	scanlines = (sl_scanline **) malloc( sizeof(sl_scanline *) * yres); 
	nsc = NULL;
	sl_heap = NULL;
	for (y = yres - 1 ; y >= 0 ; y-- ) {
		sc = new sl_scanline;
		scanlines[y] = sc;
		scanlines[y]->index = y;
		sc->unsorted = NULL;
		sc->sorted = NULL;
		sc->unsorted_tail = NULL;
		sc->sorted_tail = NULL;
		scanlines[y]->next = nsc;
		nsc = sc;
	}
}


///////
///////  Functions:
///////

void sl_screen::sl_screen_add_unsorted( int iscanline, 
        int i_in_x, int i_out_x, float inz, float outz, itriangle *itri)

// places node at beginning of list
{
	sl_node *tempnode, *newnode, *heapnode;

	tempnode = scanlines[iscanline]->unsorted;
	// if empty, make new one
	if (tempnode == NULL) {
		if (sl_heap == NULL) {	
			tempnode = new sl_node;
		} else {
			tempnode = sl_heap;
			sl_heap = sl_heap->next;
			sl_used_count++;
		}
		tempnode->next = NULL;
		scanlines[iscanline]->unsorted_tail = tempnode;
		newnode = tempnode;
	} else {
		if (sl_heap == NULL)
			newnode = new sl_node;
		else {
			newnode = sl_heap;
			sl_heap = sl_heap->next;
			sl_used_count++;
		}
		newnode->next = tempnode;
	}

	if (i_in_x <= i_out_x)  {
		newnode->startpixel = i_in_x;
		newnode->endpixel = i_out_x;
	} else {
		newnode->startpixel = i_out_x;
		newnode->endpixel = i_in_x;
	}
	newnode->zin = inz;
	newnode->zout = outz;
        newnode->tri = itri;
	newnode->zdelta = 0;
	scanlines[iscanline]->unsorted = newnode;
	sl_node_count++;
}
void sl_screen::sl_screen_add_sorted( int iscanline, 
        int i_in_x, int i_out_x, float inz, float outz, itriangle *itri)

// places node at beginning of list
{
	sl_node *tempnode, *newnode, *heapnode;

	tempnode = scanlines[iscanline]->sorted;
	// if empty, make new one
	if (tempnode == NULL) {
		if (sl_heap == NULL) {	
			tempnode = new sl_node;
		} else {
			tempnode = sl_heap;
			sl_heap = sl_heap->next;
			sl_used_count++;
		}
		tempnode->next = NULL;
		scanlines[iscanline]->sorted_tail = tempnode;
		newnode = tempnode;
	} else {
		if (sl_heap == NULL)
			newnode = new sl_node;
		else {
			newnode = sl_heap;
			sl_heap = sl_heap->next;
			sl_used_count++;
		}
		newnode->next = tempnode;
	}

	newnode->startpixel = i_in_x;
	newnode->endpixel = i_out_x;
	newnode->zin = inz;
	newnode->zout = outz;
        newnode->tri = itri;
	scanlines[iscanline]->sorted = newnode;
	sl_node_count++;
}

void sl_screen::free_unsorted_sl_nodes(sl_scanline *sc)
{

	sl_node *p;

	if ( sc->unsorted != NULL) {
		p = sl_heap;
		sl_heap = sc->unsorted;
		sc->unsorted = NULL;
		sc->unsorted_tail->next = p;
		sc->unsorted_tail = NULL;
	}
}

void sl_screen::free_sorted_sl_nodes()
{

	int y;
	sl_scanline *sc;
	sl_node *p;

	for (y = 0 ; y < yres ; y++ ) {
		sc = scanlines[y];
		if ( sc->sorted != NULL) {
			p = sl_heap;
			sl_heap = sc->sorted;
			sc->sorted = NULL;
			sc->sorted_tail->next = p;
			sc->sorted_tail = NULL;
		}
	}
			
}

/* 
// great candidate for fixed point
// since we're so close to integers.
FUNCTION: get_scanlines( itriangle *tri, int xres, int yres)

	WARNING: _most_ of this code is sorting and clipping.
		 and only emitting stuff on the screen.

This function splits a triangle into scanlines and
builds an unsorted list of possibly overlapping segments,
one list per scanline.

	sort points of triangle to look like:

	1)				2)

		a        -=OR=-         a
	    __/ |		        |\
	  _/     |			| \
	b         |			|  >b
	 -------   |			| /
	        --- c			|/
					c
	
	(two cases, b on left or right.  a is always on top,
					 c is always on bottom.
	determined by the cross product of vec(a,b) and vec(a,c)
	one indicates clockwise (b on right) or counter clockwise
	(b on left))

	Then bound triangle against screen, and quit early if triangle
	misses screen. 

	calculate slopes of ab, bc, ac ; use for fast scanline calc.
	denote slopes sab, sbc, sac, and scale so slope = delta x / y=1,
	e.g. slope is change in x as you go down one scanline.

	the scanlines endpoints are [ in_x -- out_x ].  When the scanline
	of in_x and out_x hits point b, either in_x or out_x is re-initialized
	to b.x, depending on type 1 or type 2.

	Substatial weird clipping goes on, so, for instance, the following case
	is trickey:


		c
	       / \
              /   \
	     /     \    <- off screen
            b       \
              \      \
     +==========:=====\===============+   <-- scanline 0 (top of screen)
     |           \     \              |
     |             \    \             |
     |               \   \            |
     |                 \  \           |  <- screen
     |                   \ \          |
     |                     \\         |
     |                       a        |
	   clipping med_x to so it happens at
		the ":"  requires knowing if b < 0.  If you read
		this code, you'll see how this is done.
		( heh, no thanks! =-)
	
*/
	

void get_scanlines(sl_screen *screen, itriangle *tri, int xres, int yres)
{
	d2 a, b, c, t;
	int sl, sab, sac, sbc, dab, dac, dbc, xab, xbc, xac, 
		bounded_ay, bounded_by, bounded_cy,
		leftedge, rightedge, tempint,
		ab_cross_ac, scanline, in_x, out_x, med_x;
	real a_z, b_z, c_z, zab, zbc, zac, in_z, out_z, med_z, tempfloat;

	a = tri->a->project;
	b = tri->b->project;
	c = tri->c->project;
	
	sab = sbc = sac = 0;
	zab = zbc = zac = 0.0;


	leftedge = a.x ; rightedge = a.x;
	if (b.x < leftedge)
		leftedge = b.x;
	if (c.x < leftedge)
		leftedge = c.x;
	if (b.x > rightedge)
		rightedge = b.x;
	if (c.x > rightedge)
		rightedge = c.x;
	// don't bother drawing if the triangle is totally to the left or
	// totally to the right of the screen:
	if ((leftedge > xres) || (rightedge < 0))
		return;

	//debug printf("ABC before sort: a: %d %d b: %d, %d, c: %d %d \n",
		//debug a.x, a.y, b.x, b.y, c.x, c.y);

	
	// sort a, b, c by scanline (y value)
	// and get sorted 2d integer points a, b, c and their floating point depths
	if (a.y  > b.y)
	{
		if (a.y > c.y)
		{
		// a is biggest
			if (b.y > c.y)
			{ // c is smallest 	(a b c) 
			a_z = tri->a->trans->z;
			b_z = tri->b->trans->z;
			c_z = tri->c->trans->z;
			}
			else 
			{ // b is smallest 	(a c b)
			t.x = c.x; t.y = c.y;
			c.x = b.x; c.y = b.y;
			b.x = t.x; b.y = t.y; 
			a_z = tri->a->trans->z;
			c_z = tri->b->trans->z;
			b_z = tri->c->trans->z;
			} 
		} else {
			// c is biggest 		(c a b)
			t.x = b.x; t.y = b.y;
			b.x = a.x; b.y = a.y;
			a.x = c.x; a.y = c.y;
			c.x = t.x; c.y = t.y;
			c_z = tri->a->trans->z;
			a_z = tri->b->trans->z;
			b_z = tri->c->trans->z;
		}
	} else {
		if (b.y > c.y)
		{
		// b is biggest
			if (a.y > c.y)
			{
			// c is smallest 	(b a c)
			t.x = a.x; t.y = a.y;
			a.x = b.x; a.y = b.y;
			b.x = t.x; b.y = t.y;
			b_z = tri->a->trans->z;
			a_z = tri->b->trans->z;
			c_z = tri->c->trans->z;
			}
			else {
			// a is smallest 	(b c a)
			t.x = b.x; t.y = b.y;
			b.x = c.x; b.y = c.y;
			c.x = a.x; c.y = a.y;
			a.x = t.x; a.y = t.y;
			b_z = tri->a->trans->z;
			c_z = tri->b->trans->z;
			a_z = tri->c->trans->z;
			}
		} else {
		// c is biggest 		(c b a)
		t.x = c.x; t.y = c.y;
		c.x = a.x; c.y = a.y;
		a.x = t.x; a.y = t.y; 
			c_z = tri->a->trans->z;
			b_z = tri->b->trans->z;
			a_z = tri->c->trans->z;
		}
		
	} // end [if(a.y>b.y)]
	// don't bother drawing if the triangle is totally above or
	// totally below the screen:
	if ((a.y <= 0) || (c.y >= yres))
		return;
	if ((a_z <= 0.0) && (b_z <= 0.0) && (c_z <= 0.0))
		return;

	ab_cross_ac = (a.x - b.x) * (a.y - c.y) - (a.y - b.y) * (a.x - c.x);
	dab = a.y - b.y; dac = a.y - c.y; dbc = b.y - c.y;
	xab = a.x - b.x; xac = a.x - c.x; xbc = b.x - c.x;
	if (dab != 0) {
		sab =  ( xab << FIXED_SHIFT_EXP) / dab;
		zab = (a_z - b_z) / dab;
	}
	if (dbc != 0) {
		sbc = (xbc << FIXED_SHIFT_EXP) / dbc;
		zbc = (b_z - c_z) / dbc;
	}
	if (dac != 0) {
		sac =  (xac << FIXED_SHIFT_EXP) / dac;
		zac = (a_z - c_z) / dac;
	}
//debug 	printf("Triangle stats: a (%d %d %3.2f) b(%d %d %3.2f) c(%d %d %3.2f)  cross: %d\n",
//debug 				a.x, a.y, a_z,  b.x, b.y, b_z,  c.x, c.y, c_z , ab_cross_ac);
//debug 	printf("              : sab (%d) = 1K * dab(%d) / xab(%d) \n",sab,dab,xab);
//debug 	printf("              : sbc (%d) = 1K * dbc(%d) / xbc(%d) \n",sbc,dbc,xbc);
//debug 	printf("              : sac (%d) = 1K * dac(%d) / xac(%d) \n",sac,dac,xac);

	// start at point c and sweep to point a with scanlines.
	
	// Assume type 1.  At end, check for type 2 and
	// swap end pts and z's.
 
	// Hassan Chop!  Yucky! This is just clipping the triangle
	// against the screen.
	if (b.y < 0) {
		//  a.y >= 0 >  b.y > c.y
		//bounded_cy = (c.y < 0) ? 0 : c.y;
		bounded_cy =  0;
		//bounded_by = (b.y > yres) ? yres : b.y;
		//bounded_by = (b.y < 0) ? 0: bounded_by;
		bounded_by =  0;
		bounded_ay = (a.y > yres) ? yres : a.y;
		in_x =  b.x  -   ((  b.y * sab ) >> FIXED_SHIFT_EXP);
		out_x = c.x  -   (( c.y * sac ) >> FIXED_SHIFT_EXP);
		in_z = - b.y * zab + b_z;
		out_z = -c.y * zac + c_z;
		med_x = in_x << FIXED_SHIFT_EXP;
		med_z = b_z;
	}	
	else if (c.y < 0) {
		//bounded_cy = (c.y < 0) ? 0 : c.y;
		bounded_cy =  0;
		bounded_by = (b.y > yres) ? yres : b.y;
		//bounded_by = (b.y < 0) ? 0: bounded_by;
		// above impossible since b.y !< 0 from 1st if stmt.
		bounded_ay = (a.y > yres) ? yres : a.y;
		in_x =  c.x  - ((c.y * sbc ) >> FIXED_SHIFT_EXP);
		out_x = c.x  - ((c.y * sac ) >> FIXED_SHIFT_EXP);
		in_z =  c_z - c.y * zbc;
		out_z = c_z - c.y * zac;
		med_x = b.x << FIXED_SHIFT_EXP;
		med_z = b_z;
	}
	else {
		// bounded_cy = (c.y > yres) ? yres : c.y;
		// above impossible since we checked for off screen way above
		bounded_cy = c.y;
		bounded_by = (b.y > yres) ? yres : b.y;
		bounded_ay = (a.y > yres) ? yres : a.y;
		in_x = out_x = c.x;
		in_z = out_z = c_z;
		med_z = b_z;
		med_x = b.x << FIXED_SHIFT_EXP;
	}
	if (ab_cross_ac <= 0)  {
		// Type 2, swap in and out
		tempint = in_x;
		in_x = out_x;
		out_x = tempint;
		tempfloat = in_z;
		in_z = out_z;
		out_z = tempfloat;
	}
	in_x <<= FIXED_SHIFT_EXP;
	out_x <<= FIXED_SHIFT_EXP;

	if (in_x > out_x) {
		printf("What?! in_x %d > out_x %d \n", in_x, out_x);
		printf("Triangle stats: a (%d %d %3.2f) b(%d %d %3.2f) c(%d %d %3.2f)  cross: %d\n",
					a.x, a.y, a_z,  b.x, b.y, b_z,  c.x, c.y, c_z , ab_cross_ac);
		printf("              : sab (%d) = 1K * dab(%d) / xab(%d) \n",sab,dab,xab);
		printf("              : sbc (%d) = 1K * dbc(%d) / xbc(%d) \n",sbc,dbc,xbc);
		printf("              : sac (%d) = 1K * dac(%d) / xac(%d) \n",sac,dac,xac);
		//exit(0);
		return;
	}

	if (ab_cross_ac > 0)  	// Type 1
	{
		//debug printf("type 1\n");
		for (sl = bounded_cy; sl < bounded_by ; sl++ )
		{
			in_x += sbc;
			out_x += sac;
			in_z  += zbc;
			out_z += zac;
			//emit_stub(sl, in_x, out_x, xres);
			emit_segment(screen, sl, in_x >> FIXED_SHIFT_EXP, out_x >> FIXED_SHIFT_EXP, 
					xres, in_z, out_z, tri);
		}
		//debug printf("type 11\n");
		in_x = med_x;
		in_z = med_z;
		for (sl = bounded_by; sl < bounded_ay ; sl++ )
		{
			in_x += sab;
			out_x += sac;
			in_z  += zab;
			out_z += zac;
			//emit_stub(sl, in_x, out_x, xres);
			emit_segment(screen, sl, in_x >> FIXED_SHIFT_EXP, out_x >> FIXED_SHIFT_EXP, 
					xres, in_z, out_z, tri);
		}
	
	} else {		// Type 2
		//debug printf("type 2 bcy %d < bby %d ++\n", bounded_cy, bounded_by);
		for (sl = bounded_cy; sl < bounded_by ; sl++ )
		{
			in_x += sac;
			out_x += sbc;
			in_z  += zac;
			out_z += zbc;
			//emit_stub(sl, in_x, out_x, xres);
			emit_segment(screen, sl, in_x >> FIXED_SHIFT_EXP, out_x >> FIXED_SHIFT_EXP, 
					xres, in_z, out_z, tri);
		}
		//debug printf("type 22\n");
		out_x = med_x;
		out_z = med_z;
		for (sl = bounded_by; sl < bounded_ay ; sl++ )
		{
			in_x += sac;
			out_x += sab;
			in_z  += zac;
			out_z += zab;
			//emit_stub(sl, in_x, out_x, xres);
			emit_segment(screen, sl, in_x >> FIXED_SHIFT_EXP, out_x >> FIXED_SHIFT_EXP, 
					xres, in_z, out_z, tri);
		}
	}
}

void emit_stub(int scanline, int in_x, int out_x, int xres)
{
	int i_in_x, i_out_x;

	//printf("seg %3d: [%3d, %3.2f] ... [%3d, %3.2f]\n", scanline,
		//in_x / 1000, in_z, out_x / 1000, out_z);
	if ((in_x < xres << FIXED_SHIFT_EXP) && (out_x > 0))
	{
		i_in_x = (in_x < 0)? 0 : in_x >> FIXED_SHIFT_EXP;
		i_out_x = (out_x > xres << FIXED_SHIFT_EXP)? xres : out_x >> FIXED_SHIFT_EXP;
		my_vga_drawline(i_in_x, scanline, i_out_x, scanline, 10);
	}
} 
void sl_screen::draw_sorted()
{
	sl_scanline *scanline;
	sl_node *node;

	scanline = scanlines[0];
	while (scanline != NULL)
	{
		node = scanline->sorted;
		while (node != NULL)
		{
			my_vga_drawline( node->startpixel, scanline->index, 
				node->endpixel, scanline->index,
				node->tri->color);
			node = node->next;
		}
		scanline = scanline->next;
	}
}
		
void sl_screen::draw_buffer(camera *cam)
{
	unsigned char *buff, *cambuf;
	//unsigned long fourcolors;
	sl_scanline *scanline;
	sl_node *node;
	unsigned char color;
	int y, xres, x1, x2, x, slong;

	y = 0;
	
	// slong = sizeof(unsigned long);
	xres = cam->xres;
	cambuf = cam->buffer;

	scanline = scanlines[0];
	while (scanline != NULL)
	{
		node = scanline->sorted;
		while (node != NULL)
		{
			color = node->tri->color;
			//fourcolors = 0;
			//for (x = 0 ; x < slong; x++)
			//   fourcolors = fourcolors << 8 + color;
			x1 = node->startpixel;
			x2 = node->endpixel;
			buff = cambuf + y * xres + x1;
			//for (x = x1 ; x < x2 ; x += slong) {
			//	(unsigned long ) *buff =  fourcolors;
			//	(unsigned long *) buff++;
			//}
			for (x = x1 ; x < x2 ; x ++) {
				*buff =  color;
				buff += 1;
			}
			//for (x = x - x2; x > 0 ; x--)
			//	*(buff++) = color;
			node = node->next;
		}
		scanline = scanline->next;
		y++;
	}
}
void sl_screen::draw_buffer_left_right(camera *cam, char left_right)
{
	unsigned char *buff, *cambuf;
	//unsigned long fourcolors;
	sl_scanline *scanline;
	sl_node *node;
	unsigned char color;
	int y, xres, x1, x2, x, slong, l_r_mul;

	y = 0;
	
	// slong = sizeof(unsigned long);
	xres = cam->xres;
	cambuf = cam->buffer;

	if (left_right == 'l')
		l_r_mul = 0;
	else
		l_r_mul = 4;

	scanline = scanlines[0];
	while (scanline != NULL)
	{
		node = scanline->sorted;
		while (node != NULL)
		{
			if (node->tri->color != 0)
				color = (unsigned char) ( (node->tri->color >> 5) + 1 ) << l_r_mul;
			else
				color = (unsigned char) 0;
			x1 = node->startpixel;
			x2 = node->endpixel;
			buff = cambuf + y * xres + x1;
			if (left_right == 'l')
				for (x = x1 ; x < x2 ; x ++) {
					*buff =  color;
					buff += 1;
				}
			else // left_right = 'r'
				for (x = x1 ; x < x2 ; x ++) {
					*buff =  *buff | color;
					buff += 1;
				}
			node = node->next;
		}
		//scanline = scanline->next;
		scanline = scanlines[y];
		y++;
	}
}
		
void sl_screen::draw_unsorted()
{
	sl_scanline *scanline;
	sl_node *node;

	scanline = scanlines[0];
	while (scanline != NULL)
	{
		node = scanline->unsorted;
		while (node != NULL)
		{
			my_vga_drawline( node->startpixel, scanline->index, 
				node->endpixel, scanline->index,
				node->tri->color);
			node = node->next;
		}
		scanline = scanline->next;
	}
}
		
void emit_segment
(sl_screen *frozen_yogurt, int scanline, int in_x, 
        int out_x, int xres, float inz, float outz, itriangle *itri)
{
	int i_in_x, i_out_x;

	if ((in_x < xres ) && (out_x > 0))
	{
		i_in_x = (in_x < 0)? 0 : in_x ;
		i_out_x = (out_x > xres )? xres : out_x ;
		frozen_yogurt->sl_screen_add_unsorted(scanline, 
			i_in_x , i_out_x, inz, outz, itri);
	}
}


// Welcome to the slist meaty meat meat.
// This takes a sorted slist with overlapping regions and produces
// a nonoverlapping slist.  e.g.
//
//		 \
//           -----\  becomes -- --\
//             -               -
//     
// The algorithm marches through the sorted endpoints of the slist
// and keeps track of the closest slist segment.  The thought is:
// once the closest segment has been found, either the termination 
// of the current segment or a closer segment will displace the current
// closest segment.   The output segments can be entire input segments,
// bigginings or ends of input segments, or any tiny piece of an input segment.
// This is done with two sub procedures which act according to wheather or
// not the current input end point is either the head or the tail of a segment.
//
// The kahuna is the current closest segment's head.
//
//input segment endpoint p (either head or tail):
//EMIT is an output slist element with head and length as arguments
//	since each head is linked to an slist segment, it is
//	possible to get slope, etc.
//the consideration list is all the segments that we have entered
//and not left.  You can add, subtract, and find the maximum
//height and point of the consideration list. Of course, the
//max could turn out to be internal to a segment, so the max
//funtion must also return a pointer to it's parent segment
//for information like the direction.
//---------------------------------------------------------------
//head
//----
//if (kahuna == NULL)
//	kahuna->head = p
//	kahuna->segment = p->segment
//add(p->segment, consideration list)
//if |p| > max |consideration list, p.x|
//	if kahuna.x != p.x // no zero length segments, please
//		EMIT(kahuna, p.x - kahuna.x) 
//		kahuna->head = p->head
//		kahuna->segment = p->segment
//---------------------------------------------------------------
//tail
//----
//sub(p->segment, consideration list)
//if p->segment == kahuna->segment
//	EMIT(kahuna, p.x - kahuna.x)
//	kahuna->head,segment = max (consideration list)
//---------------------------------------------------------------
// 		

// from Numerical Recipes in C, ISBN 0-521-35465-X

void endpoint_heap_sort(int n, int ra[], int rb[])
{
	int l, j, ir, i, rra, rrb;

	/* l = (n >> 1) + 1; */
	l = ((n - 1) >> 1) + 1;
	ir = n - 1;
	
	for (;;) {
		if ( l > 0) {			// hiring phase
			rra = ra[--l];
			rrb = rb[l];
		} else {				// retire/promo phase
			rra = ra[ir ];
			rrb = rb[ir ];
			ra[ir ] = ra[0 ];
			rb[ir ] = rb[0 ];
			if (--ir == 0 ) {	// last promo
				ra[0 ] = rra;
				rb[0 ] = rrb;
				return;
			}
		}
		i = l;
		j = (l == 0)? 1 : (l << 1);
		while (j <= ir) {
			if ((j < ir)  && (ra[j ] < ra[j+1 ])) 
				++j;
			if (rra < ra[j ]) {
				ra[i ] = ra[j ];
				rb[i ] = rb[j ];
				i = j;
				j += j;
			}
			else j = ir + 1; 
		}
		ra[i ] = rra;
		rb[i ] = rrb;
	}
}

void endpoint_sort(sl_screen *screen)
 {
 // iterate over unsorted scanlines building sorted scanlines w/ kahuna
 	int y, count, x, yres, in_pixel, endpoint_count, *xvals, *indexes;
	int cl_list_size;
 	sl_node *unsorted_segs, *sorted_segs, **cons_list;
	sl_scanline *slscan;
	endpoint_node *endpoints, *p, kahuna;
 	bool done_insert;
	float sheight, delheight;
 
 	yres = screen->yres;
	endpoints = screen->endpoints;
	indexes = screen->endpoint_idx;
	xvals = screen->endpoint_xvals;
	cons_list = screen->consideration;
 
 	for (y = 0; y < yres; y++) {
		slscan = screen->scanlines[y];
 		unsorted_segs = slscan->unsorted;
 		sorted_segs   = slscan->sorted;
		endpoint_count = 0;
		cl_list_size   = 0;
		kahuna.segment = NULL;
		// fill endpoint array for sorting
		// the "<<" jazz is to make sure that the head is always
		// sorted before the tail, in case they have the same xval
		while (unsorted_segs != NULL) {
			endpoints[endpoint_count].head_or_tail = true;
			endpoints[endpoint_count].xpixel =
				unsorted_segs->startpixel;
			xvals[endpoint_count] =
				unsorted_segs->startpixel << 1;
			endpoints[endpoint_count].segment =
				unsorted_segs;
			indexes[endpoint_count] = endpoint_count;
			endpoint_count++;
			endpoints[endpoint_count].head_or_tail = false;
			endpoints[endpoint_count].xpixel =
				unsorted_segs->endpixel;
			xvals[endpoint_count] =
				(unsorted_segs->endpixel << 1) + 1;
			endpoints[endpoint_count].segment =
				unsorted_segs;
			indexes[endpoint_count] = endpoint_count;
			endpoint_count++;
			unsorted_segs = unsorted_segs->next;
		}
		// sort endpoint array
		if (endpoint_count > 0) 
			endpoint_heap_sort(endpoint_count, xvals, indexes);
		// kahuna build sorted
		for (count = 0; count < endpoint_count ; count++) {
			p = &(endpoints[indexes[count]]);
			if (p->head_or_tail) {                        // head
				consideration_list_add(cl_list_size, cons_list, 
					p->segment);
				cl_list_size++;
				if (kahuna.segment == NULL) {
					kahuna.xpixel = p->xpixel;
					kahuna.segment = p->segment;
				} else {
					sheight = segheight(kahuna.segment, p->xpixel);
					delheight = sheight - p->segment->zin;
					delheight *= (delheight < 0.0) ? - 1.0 : 1.0;
					if ((p->segment->zin < sheight) && (delheight > Z_CLOSE)) {
						if (p->xpixel != kahuna.xpixel)
						   screen->sl_screen_add_sorted(y, 
					 		kahuna.xpixel, p->xpixel, 
							0.0, 0.0,
					 		kahuna.segment->tri);
						kahuna.xpixel = p->xpixel;
						kahuna.segment = p->segment;
					}
					if (delheight <= Z_CLOSE) {
						if (segheight(kahuna.segment, p->xpixel + 4) > 
						    segheight(p->segment, p->xpixel + 4)) 	{
							if (p->xpixel != kahuna.xpixel)
								screen->sl_screen_add_sorted(y, 
					 			kahuna.xpixel, p->xpixel, 
								0.0, 0.0,
					 			kahuna.segment->tri);
							kahuna.xpixel = p->xpixel;
							kahuna.segment = p->segment;
						}
					}
				}
			} else {					// tail
				consideration_list_remove(cl_list_size, cons_list,
					p->segment);
				cl_list_size--;
				if (p->segment == kahuna.segment)  {
					// add sorted
					screen->sl_screen_add_sorted(y, 
					 	kahuna.xpixel, p->xpixel, 
						0.0, 0.0,
					 	kahuna.segment->tri);
					kahuna.xpixel = p->xpixel;
					kahuna.segment = consideration_maxify(
						cl_list_size, cons_list);
				}

			} // if head or tail
		} // for endpoint count
		screen->free_unsorted_sl_nodes(slscan);
	} // for all y scanlines
} // endpoint_sort
		
// add seg to con list
void consideration_list_add(int listsize, sl_node **con_list, sl_node *seg)
{
	con_list[listsize] = seg;
}
void consideration_list_remove(int listsize, sl_node **con_list, sl_node *seg)
{
	int i;

	i = 0;
	while (con_list[i] != seg)  // find seg
		i++;
	listsize--;
	while (i < listsize) {
		con_list[i] = con_list[i+1];  // move up pointers
		i++;
	}
}

sl_node * consideration_maxify(int listsize, sl_node **con_list)
{
	int i;
	float z;

	sl_node *max;

	i = 0;
	max = NULL;
	z = 100000;
	while (i < listsize) {
		if (con_list[i]->zin < z) {
			max = con_list[i];
			z = max->zin;
		}
		i++;
	}	
			
	return max;
}
			
float segheight(sl_node *seg, int x)
{
    if ( x <= seg->startpixel)
	return seg->zin;
    if (seg->startpixel >= seg->endpixel)
	return seg->zout;
    if (seg->zdelta == 0)
	seg->zdelta = (seg->zout - seg->zin) / 
		(float)	(seg->endpixel - seg->startpixel);
   return (x - seg->startpixel) * seg->zdelta + seg->zin;
}
		
		 

		
				
///////
/////// Printers:
///////

sl_scanline::print()
{
	sl_node *tmp;

//	tmp = node;
//	while (tmp != NULL) {
//		tmp->print();
//		tmp = tmp->next;
//	}
}

sl_node::print()
{
	printf("[");
	if (tri == NULL)
		printf("0");
	else
	   tri->print(); 
	printf(" (%3d %3d)]", startpixel, endpixel); 
}

sl_screen::print()
{
	int  y;
	
	printf("Screen at %d has res %d, %d\n", 
		1, xres, yres);
	for( y = 0; y < yres; y++) {
		printf(" %3d ) ", scanlines[y]->index);
		scanlines[y]->print();
		printf("\n");
	}
}
